from typing import Any, Callable

from .property import ReactiveProperty


class ReactiveOwner:
    def __init__(self):
        self.__on_change_handlers = []

    def on_change(self, func: Callable[..., None], *args):
        self.__on_change_handlers.append((func, [object.__getattribute__(self, arg.name) for arg in args]))

    def __getattribute__(self, name: str):
        obj = object.__getattribute__(self, name)
        if hasattr(obj, '__get__'):
            return obj.__get__(self, type(self))
        return obj

    def _bulk_update(self, *args):
        changed = False
        funccalls = {}

        for arg in args:
            try:
                obj = object.__getattribute__(self, arg["name"])
            except AttributeError:
                object.__setattr__(self, arg["name"], arg["value"])
                if hasattr(arg["value"], '__set_name__'):
                    arg["value"].__set_name__(self, arg["name"])
            else:
                if hasattr(obj, '__set__'):
                    if isinstance(obj, ReactiveProperty):
                        prev_value = obj.value
                    obj.__set__(self, arg["value"])

                    if isinstance(obj, ReactiveProperty):
                        if prev_value != obj.value:
                            changed = True
                        for change_handler in self.__on_change_handlers:
                            if prev_value != obj.value and (obj in change_handler[1] or not change_handler[1]):
                                func = change_handler[0]
                                if func in funccalls:
                                    funccalls[func] = funccalls[func] + [obj]
                                else:
                                    funccalls[func] = [obj]
                else:
                    object.__setattr__(self, arg["name"], arg["value"])

        for func in funccalls:
            func(*funccalls[func])
        return changed

    def __setattr__(self, name: str, value: Any):
        return self._bulk_update({"name": name, "value": value})



from typing import Any, Callable, Type


class ReactiveProperty:
    def __init__(self, value: Any, name: str = None, field_type: Type[Any] = None):
        self.value = value
        self.name = name
        self.history = [value]
        self._field_type = field_type
        self._on_change_handlers = []

    def __get__(self, instance: Any, owner: Any):
        if isinstance(self.value, bool):
            class ObservableBool:
                value = self.value
                name = self.name
                history = self.history
                on_change = self.on_change
            return ObservableBool()
        else:
            class ObservableProperty(type(self.value)):
                name = self.name
                history = self.history
                on_change = self.on_change
            return ObservableProperty(self.value)

    def __setattr__(self, name: str, value: Any):
        if name == "value":
            if callable(value):
                raise TypeError("type 'function' is invalid for the value of 'ReactiveProperty'.")
            elif value is None:
                raise TypeError("value of 'ReactiveProperty' cannot be 'None'.")

        return super().__setattr__(name, value)

    def __set__(self, instance: Any, value: Any):
        if self._field_type and not isinstance(value, self._field_type):
            raise TypeError(
                f"expected an instance of type '{self._field_type.__name__}' for attribute '{self.name}', got '{type(value).__name__}' instead")

        prev = self.value
        self.value = value

        if value != prev:
            for change_handler in self._on_change_handlers:
                change_handler(self.value, prev)

    def __delete__(self, instance: Any):
        del self.value

    def __set_name__(self, owner: Any, name: str):
        self.name = name

    def on_change(self, func: Callable[[Any, Any], None]):
        self._on_change_handlers.append(func)



from typing import Any

from .owner import ReactiveOwner
from .property import ReactiveProperty


def to_tuple(val):
    if isinstance(val, list):
        return tuple(val)
    elif isinstance(val, tuple):
        return val
    elif val is None:
        return val
    else:
        return (val)


def make_all_reactive(org_cls=None, only_type=None, not_type=None):
    class AllReactive(org_cls):
        def __init__(self):
            super().__init__()
            self.__doc__ = org_cls.__doc__
            self.__name__ = org_cls.__name__
            self.__only_type = to_tuple(only_type)
            self.__not_type = to_tuple(not_type)

        def __setattr__(self, name: str, value: Any):
            if not isinstance(value, ReactiveProperty) and not name.startswith("_"):
                if self.__only_type and not isinstance(value, self.__only_type):
                    return super().__setattr__(name, value)
                if self.__not_type and isinstance(value, self.__not_type):
                    return super().__setattr__(name, value)
                try:
                    value = ReactiveProperty(value, name)
                except TypeError:
                    pass
            return super().__setattr__(name, value)
    return AllReactive


def all_reactive(org_cls=None, only_type=None, not_type=None):
    if org_cls:
        return make_all_reactive(org_cls, only_type, not_type)
    else:
        def all_reactive_wrapper(org_cls):
            return make_all_reactive(org_cls, only_type, not_type)
        return all_reactive_wrapper
